# eventBasedAnimationDemo.py
# TODO: Proper gravity and scaling

from tkinter import *
from eventBasedAnimationClass import EventBasedAnimationClass
import math
from tkinter import filedialog, messagebox

TEXT_MARGIN = 20
RADIUS = 2
COASTER_RADIUS = 5
WIDTH = 1200
HEIGHT = 600

class Coaster(object):
    def __init__(self):
        self.exists = False

    def instantiate(self, x, y, dx, dy, v):
        assert(abs((dx**2 + dy**2) - 1) < 0.0001)
        self.x0 = x
        self.y0 = y
        self.dx0 = dx
        self.dy0 = dy
        self.v0 = v

        self.x = x
        self.y = y
        self.dx = dx
        self.dy = dy
        self.v = v

        self.exists = True

    def getPosition(self):
        return (self.x, self.y)

    def reset(self):
        if self.exists:
            self.x = self.x0
            self.y = self.y0
            self.dx = self.dx0
            self.dy = self.dy0
            self.v = self.v0

class EventBasedAnimationDemo(EventBasedAnimationClass):

    def __init__(self):

        super(EventBasedAnimationDemo, self).__init__(WIDTH, HEIGHT)

        self.running = False
        self.timerDelay = 5 if self.running else None

        self.cursorPosition = (0,0)
        self.worldPosition = (0, self.height)

        self.joints = []
        self.speeds = []
        self.sections = []
        self.safes = []

        self.drawingArc = False

        self.currentGradient = None

        self.tentativeLineParameters = None
        self.tentativeLineFill = "black"

        self.initialSpeed = 0

        self.coaster = Coaster()

        self.gravity = 32

        self.trackAllowed = False

    # Controller functions

    def newJointAllowed(self, x, y):
        # Check that (x,y) are in window bounds and
        # that x is > than last x
        # (x,y) are viewport coordinates
        return x > 0 and y > 0 and x < WIDTH and y < HEIGHT and (not self.joints or x > self.joints[-1][0])

    def updateTentativeLine(self):
        if (self.drawingArc):
            if self.currentGradient != None:
                # Figure out how to connect arc 
                # Find arc that satisfies current gradient on left
                (x1,y1) = self.joints[-1]
                (x2,y2) = self.worldPosition
                y2_allowed = y1 + self.currentGradient * (x2 - x1)

                # Compute cx,cy,r that satisfies this
                # cx,cy will lie on line perpendicular to currentGradient
                # and will be equidistance from (x1,y1) and (x2,y2)
                # (will lie on line between these two points)
                # (cx,cy) is at intersection of these two points
                # Line 1: y = m1*x + c1
                # Line 2: y = m2*x + c2
                if self.currentGradient == 0:
                    cx = x1
                    m2 = -1 / (y2 - y1) * (x2 - x1)
                    c2 = (y2 + y1) / 2 - m2 * (x2 + x1) / 2
                    cy = m2 * cx + c2
                else:
                    m1 = -1 / self.currentGradient
                    c1 = y1 - m1*x1
                    if y1 == y2:
                      cx = (x1 + x2) / 2
                      cy = m1 * cx + c1
                    else:
                      m2 = -1 / (y2 - y1) * (x2 - x1)
                      c2 = (y2 + y1) / 2 - m2 * (x2 + x1) / 2
                      cx = (c2 - c1) / (m1 - m2)
                      cy = m1 * cx + c1
                r = ((cx - x1)**2 + (cy - y1)**2)**(1/2)
                self.tentativeLineParameters = (cx-r, HEIGHT-cy-r, cx+r, HEIGHT-cy+r)

                # tentativeLineStart 
                # We have currentGradient
                # tan
                self.tentativeLineStart = math.degrees(math.atan2((y1-cy), (x1-cx)))
                self.tentativeLineExtent = math.degrees(math.atan2((y2-cy), x2-cx)) - self.tentativeLineStart
                if y2 != cy:
                  self.tentativeGradient = -1 / (y2-cy) * (x2-cx)

                if self.tentativeLineStart * (self.tentativeLineStart + self.tentativeLineExtent) > 0:
                  self.tentativeLineFill = "black"
                  self.trackAllowed = True
                else:
                  self.tentativeLineFill = "red"
                  self.trackAllowed = False

        else: # drawing straight section
            if (self.joints):
                (x1,y1) = self.joints[-1]
                if self.currentGradient != None:
                    # TODO: y2_allowed should be computed as point on line closest to cursor
                    (x2,y2) = self.cursorPosition
                    m = self.currentGradient
                    (xp,yp) = (x2, HEIGHT-y2)
                    c = y1 - self.currentGradient * x1
                    x2_allowed = (m*yp + xp - m*c) / (m**2 + 1)
                    y2_allowed = y1 + m * (x2_allowed - x1)
                    self.tentativeLineParameters = (x1, HEIGHT-y1) + (x2_allowed, HEIGHT-y2_allowed)
                else:
                    self.tentativeLineParameters = (x1, HEIGHT-y1) + self.cursorPosition
                self.tentativeLineFill = "black" if self.newJointAllowed(*self.cursorPosition) else "red"

    def recalculateSpeedsAndSafety(self):
        self.speeds = [self.initialSpeed]
        self.safes = [True]
        for i in range(len(self.joints) - 1):
            v0 = self.speeds[i]
            y0 = self.joints[i][1]
            yend = self.joints[i+1][1]
            dy = y0 - yend
            v2 = v0**2 + 2*dy*self.gravity
            if v2 >= 0:
                v = math.sqrt(v2)
            else:
                v = math.nan
            self.speeds.append(v)
            self.safes.append(v0**2 > (2*yend - 2*y0) * self.gravity) # Safe if v0^2 > g(2yend - 2y0)

    def instantiateCoaster(self):
      if len(self.joints) >= 2:
        (x0,y0) = self.joints[0]
        (x1,y1) = self.joints[1]
        v = self.initialSpeed
        m = (y1 - y0) / (x1 - x0)
        dx = 1 / (1 + m**2)**0.5
        dy = m * dx
        self.coaster.instantiate(x0,y0,dx,dy,v)

    def onMousePressed(self, event):
        if self.newJointAllowed(event.x, event.y):
            if self.drawingArc:
                if self.trackAllowed: # TODO: Use proper boolean, not colour
                    # Compute current gradient
                    self.currentGradient = self.tentativeGradient
                    if (self.tentativeLineStart > 90 and self.tentativeLineStart + self.tentativeLineExtent < 90):
                      (x0,y0,x1,y1) = self.tentativeLineParameters
                      self.joints.append(((x0+x1)/2, HEIGHT-y0))
                      self.joints.append(self.worldPosition)
                      self.sections.append(("arc", (self.tentativeLineParameters, self.tentativeLineStart, 90-self.tentativeLineStart), 0))
                      self.sections.append(("arc", (self.tentativeLineParameters, 90, self.tentativeLineExtent-90+self.tentativeLineStart), self.currentGradient))
                    elif (self.tentativeLineStart < -90 and self.tentativeLineStart + self.tentativeLineExtent > -90):
                      (x0,y0,x1,y1) = self.tentativeLineParameters
                      self.joints.append(((x0+x1)/2, HEIGHT-y1))
                      self.joints.append(self.worldPosition)
                      self.sections.append(("arc", (self.tentativeLineParameters, self.tentativeLineStart, -90-self.tentativeLineStart), 0))
                      self.sections.append(("arc", (self.tentativeLineParameters, -90, self.tentativeLineExtent+90+self.tentativeLineStart), self.currentGradient))
                    else:
                      self.joints.append(self.worldPosition)
                      self.sections.append(("arc", (self.tentativeLineParameters, self.tentativeLineStart, self.tentativeLineExtent), self.currentGradient))
                      #self.drawingArc = False
            else: # Drawing straight track
                if self.tentativeLineParameters:
                    # Constrained
                    (x,y) = self.tentativeLineParameters[2:]
                    v0 = self.speeds[-1]
                    (x0,y0) = self.joints[-1]
                    self.joints.append((x, HEIGHT-y))
                else:
                    # Free
                    self.joints.append((event.x, HEIGHT-event.y))
                if (len(self.joints) >= 2):
                    (cur_x, cur_y) = self.joints[-1]
                    (prev_x, prev_y) = self.joints[-2]
                    self.currentGradient = (cur_y - prev_y) / (cur_x - prev_x)
                self.sections.append(("straight", (self.tentativeLineParameters,), self.currentGradient))
                
                # TODO: Analyze safety

            self.instantiateCoaster()
            self.recalculateSpeedsAndSafety()
        self.updateTentativeLine()
    
    def onKeyPressed(self, event):
        #print("keysym", event.keysym)
        if (event.keysym == "BackSpace"):
            if self.joints:
                self.joints.pop()
                if len(self.joints) == 0:
                    self.tentativeLineParameters = None
                self.sections.pop()
                if len(self.joints) < 2:
                    self.currentGradient = None
                else:
                    self.currentGradient = self.sections[-1][2]
            self.coaster.reset()
        elif (event.keysym in ["Right", "Left", "Up", "Down", "Tab"]):
            self.drawingArc = not self.drawingArc
        elif (event.keysym == "equal"):
            self.initialSpeed = (self.initialSpeed + 1) * 1.05
            self.coaster.v0 = self.initialSpeed
            self.coaster.reset()
        elif (event.keysym == "minus"):
            self.initialSpeed = max(self.initialSpeed / 1.05 - 1, 0)
            self.coaster.v0 = self.initialSpeed
            self.coaster.reset()
        elif (event.keysym == "space"):
            self.coaster.reset()
            self.running = not self.running
            self.timerDelay = 5 if self.running else None
            if self.timerDelay:
                self.onTimerFiredWrapper()

        self.recalculateSpeedsAndSafety()
        self.updateTentativeLine()

    def onMouseMotion(self, event):
        self.cursorPosition = (event.x, event.y)
        self.worldPosition = (event.x, self.height - event.y)

        self.updateTentativeLine()

    def evolve(self, coaster):
        pass
        # Perform runge kutta integration on coaster
        # 1. Figure out appropriate dynamics for where coaster is now
        (x,y) = coaster.getPosition()
        for i in range(len(self.joints) - 1):
            (x0,y0) = self.joints[i]
            (x1,y1) = self.joints[i+1]
            if (x0 <= x and x <= x1):
                # Figure out this track type
                (sectionType, params, m) = self.sections[i+1]
                dt = self.timerDelay/1000 * 3.5
                if sectionType == "straight":
                    # TODO: Change to runge kutta
                    # Evolve coaster according to straight dynamics
                    coaster.dx = 1 / (1 + m**2)**0.5
                    coaster.dy = m * coaster.dx

                    a = -self.gravity * m / math.sqrt(m**2 + 1)
                    vx = (coaster.v + a * dt / 2) * coaster.dx
                    vy = (coaster.v + a * dt / 2) * coaster.dy

                    coaster.x = coaster.x + vx * dt
                    coaster.y = coaster.y + vy * dt
                    coaster.v = coaster.v + a * dt
                elif sectionType == "arc":
                    # Evolve coaster according to arc dynamics
                    (x0,y0,x1,y1) = params[0]
                    cx = (x0 + x1) / 2
                    cy = HEIGHT - (y0 + y1) / 2
                    r = (x1 - x0) / 2
                    clockwise = params[2] < 0
                    # Runge Kutta
                    if clockwise:
                        vx1 = coaster.v * (coaster.y - cy) / r
                        vy1 = coaster.v * -(coaster.x - cx) / r
                        a1 = self.gravity * (coaster.x - cx) / r

                        x1 = coaster.x + vx1 * dt/2
                        y1 = coaster.y + vy1 * dt/2
                        v1 = coaster.v + a1 * dt/2

                        vx2 = v1 * (y1 - cy) / r
                        vy2 = v1 * -(x1 - cx) / r
                        a2 = self.gravity * (x1 - cx) / r

                        x2 = coaster.x + vx2 * dt/2
                        y2 = coaster.y + vy2 * dt/2
                        v2 = coaster.v + a2 * dt/2

                        vx3 = v2 * (y2 - cy) / r
                        vy3 = v2 * -(x2 - cx) / r
                        a3 = self.gravity * (x2 - cx) / r

                        x3 = coaster.x + vx3 * dt
                        y3 = coaster.y + vy3 * dt
                        v3 = coaster.v + a3 * dt

                        vx4 = v3 * (y3 - cy) / r
                        vy4 = v3 * -(x3 - cx) / r
                        a4 = self.gravity * (x3 - cx) / r

                    else:
                        vx1 = coaster.v * -(coaster.y - cy) / r
                        vy1 = coaster.v * (coaster.x - cx) / r
                        a1 = -self.gravity * (coaster.x - cx) / r

                        x1 = coaster.x + vx1 * dt/2
                        y1 = coaster.y + vy1 * dt/2
                        v1 = coaster.v + a1 * dt/2

                        vx2 = v1 * -(y1 - cy) / r
                        vy2 = v1 * (x1 - cx) / r
                        a2 = -self.gravity * (x1 - cx) / r

                        x2 = coaster.x + vx2 * dt/2
                        y2 = coaster.y + vy2 * dt/2
                        v2 = coaster.v + a2 * dt/2

                        vx3 = v2 * -(y2 - cy) / r
                        vy3 = v2 * (x2 - cx) / r
                        a3 = -self.gravity * (x2 - cx) / r

                        x3 = coaster.x + vx3 * dt
                        y3 = coaster.y + vy3 * dt
                        v3 = coaster.v + a3 * dt

                        vx4 = v3 * -(y3 - cy) / r
                        vy4 = v3 * (x3 - cx) / r
                        a4 = -self.gravity * (x3 - cx) / r

                    coaster.x = coaster.x + (vx1 + 2*vx2 + 2*vx3 + vx4) * dt / 6
                    coaster.y = coaster.y + (vy1 + 2*vy2 + 2*vy3 + vy4) * dt / 6
                    coaster.v = coaster.v + (a1 + 2*a2 + 2*a3 + a4) * dt / 6
                continue
    
    def onTimerFired(self):
        # Evolve coaster
        if self.coaster.exists:
            # Evolve coaster
            self.evolve(self.coaster)


    #
    # View functions
    # these should not update the model!
    #

    def drawPenTip(self):
        (cx,cy) = self.cursorPosition
        parameters = (cx-RADIUS, cy-RADIUS, cx+RADIUS, cy+RADIUS)
        self.canvas.create_oval(*parameters, fill="black")
        self.canvas.create_text(cx + TEXT_MARGIN, cy + TEXT_MARGIN, text = "(%d,%d)" % self.worldPosition, anchor=W)

    def drawTrackSections(self):
        for i in range(len(self.sections) - 1):
            # Connect joint i to joint i+1
            (sectionType, params, gradient) = self.sections[i+1]
            fill = "black" if self.safes[i+1] else "red"
            if sectionType == "straight":
                self.canvas.create_line(*params[0], fill=fill, width=3)
            elif sectionType == "arc":
                start = params[1]
                extent = params[2]
                self.canvas.create_arc(*params[0], style=ARC, start=start, extent=extent, outline=fill, width=3)
                '''
                # Draw centre of arc
                (x0,y0,x1,y1) = params[0]
                cx = (x0 + x1) / 2
                cy = (y0 + y1) / 2
                self.canvas.create_oval(cx-RADIUS, cy-RADIUS, cx+RADIUS, cy+RADIUS, fill="red")
                '''

    def drawJointsAndSpeeds(self):
        for i in range(len(self.joints)):
            (cx,cy) = self.joints[i]
            parameters = (cx-RADIUS, HEIGHT-cy-RADIUS, cx+RADIUS, HEIGHT-cy+RADIUS)
            self.canvas.create_oval(*parameters, fill="black")
            self.canvas.create_line(cx, 0, cx, HEIGHT, dash=(5,))

            '''
            if i < len(self.speeds):
                speed = self.speeds[i]
                text = "v = %.2f" % speed
                self.canvas.create_text(cx, TEXT_MARGIN, text=text)
            '''

    def writeInitialSpeed(self):
      if self.joints:
        (cx,cy) = self.joints[0]
        self.canvas.create_text(cx,HEIGHT-cy-TEXT_MARGIN,text="v = %.2f" % self.speeds[0])

    def drawTentativeLine(self):
        if (self.joints): # Only draw if joints isn't empty
            if (self.drawingArc):
                if self.currentGradient != None: # Only draw arc if there is a currentGradient
                    self.canvas.create_arc(*self.tentativeLineParameters, fill="", outline=self.tentativeLineFill, start=self.tentativeLineStart, extent=self.tentativeLineExtent, style=ARC)
                    # Draw rest of straight line
                    if self.tentativeLineFill == "black":
                      (x1,y1) = self.worldPosition
                      x2 = WIDTH
                      y2 = y1 + self.tentativeGradient * (x2 - x1)
                      self.canvas.create_line(x1,HEIGHT-y1,x2,HEIGHT-y2)
            else:
                self.canvas.create_line(*self.tentativeLineParameters, fill=self.tentativeLineFill)

    def writeMode(self):
        text = "Arc" if self.drawingArc else "Straight"
        self.canvas.create_text(WIDTH/2, HEIGHT - TEXT_MARGIN, text=text)

    def drawPlaybackSymbol(self):
        text = "Playing" if self.running else "Paused"
        self.canvas.create_text(WIDTH - TEXT_MARGIN, TEXT_MARGIN, text=text)

    def drawCoaster(self):
        if self.coaster.exists:
            (cx,y) = self.coaster.getPosition()
            cy = HEIGHT-y
            self.canvas.create_oval(cx-COASTER_RADIUS, cy-COASTER_RADIUS, cx+COASTER_RADIUS, cy+COASTER_RADIUS, fill="green")

    def writeCoasterSpeed(self):
        if self.coaster.exists:
            self.canvas.create_text(WIDTH - TEXT_MARGIN, HEIGHT - TEXT_MARGIN, text="%.2f"%self.coaster.v)
    
    def redrawAll(self):
        self.canvas.delete(ALL)

        # Text
        self.writeMode()

        self.drawPenTip()
        self.drawTrackSections()
        self.drawJointsAndSpeeds()
        self.drawTentativeLine()

        #self.drawPlaybackSymbol()
        self.drawCoaster()
        self.writeCoasterSpeed()
        self.writeInitialSpeed()
    
    def writeCoasterToFile(self, filePath):
        with open(filePath, "wt") as fout:
            fout.write(str((self.joints, self.sections, self.initialSpeed, self.tentativeLineParameters)))
    
    def readCoasterFromFile(self, filePath):
        try:
            with open(filePath) as fin:
                string = fin.read()
                (self.joints, self.sections, self.initialSpeed, self.tentativeLineParameters) = eval(string)
            self.recalculateSpeedsAndSafety()
            self.currentGradient = self.sections[-1][2]
            self.instantiateCoaster()
            self.redrawAll()
            return True
        except Exception as inst:
            print(inst)
            return False

    def showErrorReadingFileMessage(self):
        message = "Unable to read file!"
        title = ""
        messagebox.showwarning(title, message)
    
    def openFile(self, event=None):
        filePath = filedialog.askopenfilename(filetypes=[("RollerCoaster Tycoon X files",".rctx")])
        didOpen = self.readCoasterFromFile(filePath)
        if not didOpen:
            self.showErrorReadingFileMessage()

    def saveFile(self, event=None):
        filePath = filedialog.asksaveasfilename(defaultextension=".rctx")
        if filePath:
            self.writeCoasterToFile(filePath)
        
    def createMenu(self):
        menubar = Menu(self.root)
        filemenu = Menu(menubar)

        filemenu.add_command(label="Open", command=self.openFile, accelerator="Cmd+O")
        filemenu.add_command(label="Save", command=self.saveFile, accelerator="Cmd+S")
        menubar.add_cascade(label="File", menu=filemenu)
        self.root.config(menu = menubar)

    def initAnimation(self):
        self.root.bind("<Command-o>", self.openFile)
        self.root.bind("<Command-s>", self.saveFile)
        self.root.bind("<Command-w>", lambda event: self.root.destroy())
        self.createMenu()
        self.redrawAll()
        self.root.title("RollerCoaster Tycoon X")

EventBasedAnimationDemo().run()
